.g-job-color-fetching-input, .g-job-color-pushing-output
  background-color #89d2e2
  color #333

.g-job-color-converting-input, .g-job-color-converting-output
  background-color #92f5b5
  color white
